@@ -5,9 +5,9 @@
 #
 
 # Meta
-__version__ = '2.0'
+__version__ = '2.1'
 __license__ = "GPLv3" # see LICENSE.txt
-__version_info__ = (2, 0)
+__version_info__ = (2, 1)
 __author__ = 'Dan McDougall <daniel.mcdougall@liftoffsoftware.com>'
 
 # TODO: Add the ability to mark variables, functions, classes, and methods for non-obfuscation.
@@ -309,6 +309,9 @@ def main():
         except Exception as err:
             print("Error reading %s:" % options.prepend)
             print(err)
+    # Automatically enable --obfuscate if --nonlatin (it's implied)
+    if options.use_nonlatin:
+        options.obfuscate = True
     if len(args) > 1: # We're dealing with more than one file
         name_generator = None # So we can tell if we need to obfuscate
         if options.obfuscate or options.obf_classes \
@@ -319,7 +322,7 @@ def main():
             if options.use_nonlatin:
                 if sys.version_info[0] == 3:
                     name_generator = obfuscate.obfuscation_machine(
-                        use_unicode=True, count=identifier_length
+                        use_unicode=True, identifier_length=identifier_length
                     )
                 else:
                     print(
@@ -327,7 +330,7 @@ def main():
                     sys.exit(2)
             else:
                 name_generator = obfuscate.obfuscation_machine(
-                    count=identifier_length)
+                    identifier_length=identifier_length)
             table =[{}]
         cumulative_size = 0 # For size reduction stats
         cumulative_new = 0 # Ditto
@@ -401,7 +404,7 @@ def main():
            or options.obf_import_methods:
             identifier_length = int(options.replacement_length)
             name_generator = obfuscate.obfuscation_machine(
-                count=identifier_length)
+                identifier_length=identifier_length)
             obfuscate.obfuscate(module, tokens, options)
         # Convert back to text
         result = token_utils.untokenize(tokens)
@@ -412,10 +415,12 @@ def main():
             result = compression.gz_pack(result)
         elif lzma and options.lzma:
             result = compression.lzma_pack(result)
-        result += "\n# Created by pyminifier.py\n"
+        result += (
+                "# Created by pyminifier "
+                "(https://github.com/liftoff/pyminifier)\n")
         # Either save the result to the output file or print it to stdout
         if options.outfile:
-            f = open(options.outfile, 'w')
+            f = open(options.outfile, 'w', encoding='utf-8')
             f.write(result)
             f.close()
             new_filesize = os.path.getsize(options.outfile)
